package temp;

class App {
    public static void main(final String[] ar) {
        System.out.println();
        System.currentTimeMillis();
        System.currentTimeMillis();

        System.class.getClass();

        ar.toString();
        System.class.isArray();
    }

    public static void demo() {
    }
}
